/* Nokia 5110 LCD AVR Library example
 *
 * Copyright (C) 2015 Sergey Denisov.
 * Written by Sergey Denisov aka LittleBuster (DenisovS21@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version 3
 * of the Licence, or (at your option) any later version.
 *
 * Original library written by SkewPL, http://skew.tk
 */

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>

#include "nokia5110.h"

#define UP 1
#define RIGHT 2
#define DOWN 3
#define LEFT 4
#define SNAKE_MAX 100

struct point {
	uint8_t x;
	uint8_t y;
};

void print_snake(struct point *snake, int head, int tail, struct point food)
{

	int i;
	nokia_lcd_clear();
	i = tail;
	while (1) {
		nokia_lcd_set_cursor(snake[i].x, snake[i].y);
		nokia_lcd_write_char('*', 1);
		if (i == head)
			break;
		i = (i + 1) % SNAKE_MAX;
	}
	nokia_lcd_set_cursor(food.x, food.y);
	nokia_lcd_write_char('+', 1);
	nokia_lcd_render();
}

void generate_food(struct point *snake, int head, int tail, struct point *food)
{
	int i, good;
	while (1) {
		while (1) {
			food->x = rand() % 81;
			if (food->x % 4 == 0)
				break;
		}
		while (1) {
			food->y = 15 + rand() % 26;
			if (food->y % 5 == 0)
				break;
		}
		good = 1;
		i = tail;
		while (1) {
			if (food->x == snake[i].x && food->y == snake[i].y)
				good = 0;
			if (i == head)
				break;
			i = (i + 1) % SNAKE_MAX;
		}
		if (good == 1)
			break;
	}
}

void init_pins()
{
	// LEFT
	DDRD &= ~(1 << PD0);
	PORTD |= (1 << PD0);
	// UP
	DDRD &= ~(1 << PD1);
	PORTD |= (1 << PD1);
	// RIGHTx
	DDRD &= ~(1 << PD4);
	PORTD |= (1 << PD4);
	// DOWN
	DDRD &= ~(1 << PD5);
	PORTD |= (1 << PD5);
	//ENTER
	DDRD &= ~(1 << PD6);
	PORTD |= (1 << PD6);
	//LED1
	DDRC |= (1 << PC0);
	PORTC &= ~(1 << PC0);
	//LED2
	DDRC |= (1 << PC1);
	PORTC &= ~(1 << PC1);
	//LED3
	DDRC |= (1 << PC2);
	PORTC &= ~(1 << PC2);
	//BUZZER
	DDRC |= (1 << PC3);
	PORTC &= ~(1 << PC3);
}
void init_snake(struct point *snake)
{
	uint8_t i;
	struct point poz;
	poz.y = 20;
	for (i = 0; i < 3; i++) {
		poz.x = i * 4;
		snake[i] = poz;
	}
}
void print_score(int score)
{
	char buffer[10];
	nokia_lcd_set_cursor(0, 0);
	nokia_lcd_write_string("SCORE:", 1);
	nokia_lcd_write_string(itoa(score, buffer, 10), 1);
	nokia_lcd_render();
}
void run_game(int speed)
{
	uint8_t direction = RIGHT;
	struct point snake[SNAKE_MAX];
	struct point food;
	int head = 2;
	int tail = 0;
	int last_head, i, got_input;
	int score = 0;
	init_snake(snake);
	generate_food(snake, head, tail, &food);
	while (1) {
		print_snake(snake, head, tail, food);
		print_score(score);
		got_input = 0;
		for (i = 0; i < speed; i++) {
			if ((PIND & (1 << PD0)) == 0 && direction != RIGHT && got_input == 0) {
				direction = LEFT;
				got_input = 1;
			}
			if ((PIND & (1 << PD1)) == 0 && direction != DOWN && got_input == 0) {
				direction = UP;
				got_input = 1;
			}
			if ((PIND & (1 << PD4)) == 0 && direction != LEFT && got_input == 0) {
				direction = RIGHT;
				got_input = 1;

			}
			if ((PIND & (1 << PD5)) == 0 && direction != UP && got_input == 0) {
				direction = DOWN;
				got_input = 1;
			}
			if ((PIND & (1 << PD6)) == 0) {
				_delay_ms(300);
				nokia_lcd_set_cursor(55, 0);
				nokia_lcd_write_string("PAUSE", 1);
				nokia_lcd_render();
				while (1)
					if ((PIND & (1 << PD6)) == 0)
						break;
				print_snake(snake, head, tail, food);
				print_score(score);
				_delay_ms(300);
			}
			_delay_ms(10);
		}
		last_head = head;
		head = (head + 1) % SNAKE_MAX;
		switch (direction) {
		case(RIGHT):
			if (snake[last_head].x + 4 > 80)
				snake[head].x = 0;
			else
				snake[head].x = snake[last_head].x + 4;
			snake[head].y = snake[last_head].y;
			break;
		case(LEFT):
			if (snake[last_head].x >= 4)
				snake[head].x = snake[last_head].x - 4;
			else
				snake[head].x = 80;
			snake[head].y = snake[last_head].y;
			break;
		case(DOWN):
			snake[head].x = snake[last_head].x;
			if (snake[last_head].y + 5 > 40)
				snake[head].y = 15;
			else
				snake[head].y = snake[last_head].y + 5;
			break;
		case(UP):
			snake[head].x = snake[last_head].x;
			if (snake[last_head].y >= 15)
				snake[head].y = snake[last_head].y - 5;
			else
				snake[head].y = 40;
			break;
		}
		if (snake[head].x == food.x && snake[head].y == food.y) {
			generate_food(snake, head, tail, &food);
			PORTC ^= (1 << PC0) ^ (1 << PC3);
			_delay_ms(25);
			PORTC ^= (1 << PC0) ^ (1 << PC3);
			score++;
		}
		else
			tail = (tail + 1) % SNAKE_MAX;
		i = tail;
		while (1) {
			if (i == head)
				break;
			if(snake[i].x == snake[head].x && snake[i].y == snake[head].y)
				return;
			i = (i + 1) % SNAKE_MAX;
		}
	}

}
void print_start(int speed)
{
	int i;
	nokia_lcd_clear();
	nokia_lcd_set_cursor(10, 0);
	nokia_lcd_write_string("SNAKE!", 2);
	nokia_lcd_set_cursor(10, 35);
	nokia_lcd_write_string("Press ENTER!", 1);
	nokia_lcd_set_cursor(0, 22);
	nokia_lcd_write_string("SPEED [      ]", 1);
	nokia_lcd_set_cursor(42, 22);
	for (i = 50; i >= speed; i -= 5)
		nokia_lcd_write_string("+", 1);
	nokia_lcd_render();

}
void start_game(unsigned int *seed, int *speed)
{
	print_start(*speed);
	_delay_ms(500);
	while (1) {
		if ((PIND & (1 << PD6)) == 0)
			break;
		if ((PIND & (1 << PD0)) == 0 && *speed < 50) {
			*speed += 5;
			_delay_ms(50);
		}
		if ((PIND & (1 << PD4)) == 0 && *speed > 25) {
			*speed -= 5;
			_delay_ms(50);
		}
		print_start(*speed);
		(*seed)++;
	}
	_delay_ms(200);
}
void game_over(void)
{	
	nokia_lcd_clear();
	nokia_lcd_set_cursor(20, 0);
	nokia_lcd_write_string("DEAD", 2);
	nokia_lcd_set_cursor(10, 20);
	nokia_lcd_write_string("Press Enter!", 1);
	nokia_lcd_render();
	PORTC |= (1 << PC3);
	_delay_ms(500);
	PORTC &= (0 << PC3);
	while (1) {
		PORTC ^= (1 << PC0);
		_delay_ms(100);
		PORTC ^= (1 << PC1);
		_delay_ms(100);
		PORTC ^= (1 << PC2);
		_delay_ms(100);
		if ((PIND & (1 << PD6)) == 0) {
			PORTC &= (0 << PC0);
			PORTC &= (0 << PC1);
			PORTC &= (0 << PC2);
			break;
		}
	}
}
int main(void)
{
	unsigned int seed = 0;
	int speed = 50;
    nokia_lcd_init();
	init_pins();
    for (;;) {
		start_game(&seed, &speed);
		srand(seed);
		run_game(speed);
		game_over();
    }
}
